import pytest
from proto_schema_parser import ast

from recap.converters.protobuf import ProtobufConverter
from recap.types import (
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    StringType,
    StructType,
    UnionType,
)


def test_to_recap_converter():
    protobuf_schema = """
    syntax = "proto3";
    message Test {
        string name = 1;
        bool is_valid = 2;
        bytes data = 3;
        int32 id = 4;
        int64 large_id = 5;
        uint32 uid = 6;
        uint64 large_uid = 7;
        sint32 sid = 8;
        sint64 large_sid = 9;
        float score = 10;
        double large_score = 11;
        fixed32 fixed_id = 12;
        fixed64 large_fixed_id = 13;
        sfixed32 sfixed_id = 14;
        sfixed64 large_sfixed_id = 15;
    }
    """
    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert isinstance(recap_schema, StructType)
    assert recap_schema.alias == "_root.Test"
    fields = recap_schema.fields
    assert len(fields) == 15

    # Field 1: string name = 1;
    assert isinstance(fields[0], UnionType)
    assert fields[0].extra_attrs["name"] == "name"
    assert fields[0].extra_attrs["default"] == None
    assert isinstance(fields[0].types[1], StringType)
    assert fields[0].types[1].bytes_ == 2_147_483_648
    assert fields[0].types[1].variable == True

    # Field 2: bool is_valid = 2;
    assert isinstance(fields[1], UnionType)
    assert fields[1].extra_attrs["name"] == "is_valid"
    assert fields[1].extra_attrs["default"] == None
    assert isinstance(fields[1].types[1], BoolType)

    # Field 3: bytes data = 3;
    assert isinstance(fields[2], UnionType)
    assert fields[2].extra_attrs["name"] == "data"
    assert fields[2].extra_attrs["default"] == None
    assert isinstance(fields[2].types[1], BytesType)
    assert fields[2].types[1].bytes_ == 2_147_483_648
    assert fields[2].types[1].variable == True

    # Field 4: int32 id = 4;
    assert isinstance(fields[3], UnionType)
    assert fields[3].extra_attrs["name"] == "id"
    assert fields[3].extra_attrs["default"] == None
    assert isinstance(fields[3].types[1], IntType)
    assert fields[3].types[1].bits == 32
    assert fields[3].types[1].signed == True

    # Field 5: int64 large_id = 5;
    assert isinstance(fields[4], UnionType)
    assert fields[4].extra_attrs["name"] == "large_id"
    assert fields[4].extra_attrs["default"] == None
    assert isinstance(fields[4].types[1], IntType)
    assert fields[4].types[1].bits == 64
    assert fields[4].types[1].signed == True

    # Field 6: uint32 uid = 6;
    assert isinstance(fields[5], UnionType)
    assert fields[5].extra_attrs["name"] == "uid"
    assert fields[5].extra_attrs["default"] == None
    assert isinstance(fields[5].types[1], IntType)
    assert fields[5].types[1].bits == 32
    assert fields[5].types[1].signed == False

    # Field 7: uint64 large_uid = 7;
    assert isinstance(fields[6], UnionType)
    assert fields[6].extra_attrs["name"] == "large_uid"
    assert fields[6].extra_attrs["default"] == None
    assert isinstance(fields[6].types[1], IntType)
    assert fields[6].types[1].bits == 64
    assert fields[6].types[1].signed == False

    # Field 8: sint32 sid = 8;
    assert isinstance(fields[7], UnionType)
    assert fields[7].extra_attrs["name"] == "sid"
    assert fields[7].extra_attrs["default"] == None
    assert isinstance(fields[7].types[1], IntType)
    assert fields[7].types[1].bits == 32
    assert fields[7].types[1].signed == True

    # Field 9: sint64 large_sid = 9;
    assert isinstance(fields[8], UnionType)
    assert fields[8].extra_attrs["name"] == "large_sid"
    assert fields[8].extra_attrs["default"] == None
    assert isinstance(fields[8].types[1], IntType)
    assert fields[8].types[1].bits == 64
    assert fields[8].types[1].signed == True

    # Field 10: float score = 10;
    assert isinstance(fields[9], UnionType)
    assert fields[9].extra_attrs["name"] == "score"
    assert fields[9].extra_attrs["default"] == None
    assert isinstance(fields[9].types[1], FloatType)
    assert fields[9].types[1].bits == 32

    # Field 11: double large_score = 11;
    assert isinstance(fields[10], UnionType)
    assert fields[10].extra_attrs["name"] == "large_score"
    assert fields[10].extra_attrs["default"] == None
    assert isinstance(fields[10].types[1], FloatType)
    assert fields[10].types[1].bits == 64

    # Field 12: fixed32 fixed_id = 12;
    assert isinstance(fields[11], UnionType)
    assert fields[11].extra_attrs["name"] == "fixed_id"
    assert fields[11].extra_attrs["default"] == None
    assert isinstance(fields[11].types[1], IntType)
    assert fields[11].types[1].bits == 32
    assert fields[11].types[1].signed == False

    # Field 13: fixed64 large_fixed_id = 13;
    assert isinstance(fields[12], UnionType)
    assert fields[12].extra_attrs["name"] == "large_fixed_id"
    assert fields[12].extra_attrs["default"] == None
    assert isinstance(fields[12].types[1], IntType)
    assert fields[12].types[1].bits == 64
    assert fields[12].types[1].signed == False

    # Field 14: sfixed32 sfixed_id = 14;
    assert isinstance(fields[13], UnionType)
    assert fields[13].extra_attrs["name"] == "sfixed_id"
    assert fields[13].extra_attrs["default"] == None
    assert isinstance(fields[13].types[1], IntType)
    assert fields[13].types[1].bits == 32
    assert fields[13].types[1].signed == True

    # Field 15: sfixed64 large_sfixed_id = 15;
    assert isinstance(fields[14], UnionType)
    assert fields[14].extra_attrs["name"] == "large_sfixed_id"
    assert fields[14].extra_attrs["default"] == None
    assert isinstance(fields[14].types[1], IntType)
    assert fields[14].types[1].bits == 64
    assert fields[14].types[1].signed == True


def test_to_recap_converter_repeated():
    protobuf_schema = """
        syntax = "proto3";
        message Test {
            repeated int32 values = 1;
        }
    """

    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    ListType(IntType(bits=32, signed=True)),
                ],
                name="values",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_to_recap_converter_map():
    protobuf_schema = """
        syntax = "proto3";
        message Test {
            map<string, int32> value = 1;
        }
    """

    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert isinstance(recap_schema, StructType)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    MapType(
                        StringType(variable=True, bytes_=2_147_483_648),
                        IntType(bits=32, signed=True),
                    ),
                ],
                name="value",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_to_recap_converter_forward_reference():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            Inner outerValue = 1;
        }

        message Inner {
            int32 innerValue = 1;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)

    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="_root.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="_root.Outer",
    )


def test_to_recap_converter_map_forward_reference():
    protobuf_schema = """
        syntax = "proto3";
        message Test {
            map<string, Inner> testValue = 1;
        }

        message Inner {
            int32 innerValue = 1;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)

    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    MapType(
                        StringType(variable=True, bytes_=2_147_483_648),
                        StructType(
                            [
                                UnionType(
                                    [
                                        NullType(),
                                        IntType(bits=32, signed=True),
                                    ],
                                    name="innerValue",
                                    default=None,
                                )
                            ],
                            alias="_root.Inner",
                        ),
                    ),
                ],
                name="testValue",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_to_recap_converter_nested_message():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            message Inner {
                int32 innerValue = 1;
            }
            Inner outerValue = 2;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="_root.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="_root.Outer",
    )


def test_to_recap_converter_enum():
    protobuf_schema = """
        syntax = "proto3";
        message WeatherReport {
            Season season = 1;
        }
        enum Season {
            SPRING = 0;
            SUMMER = 1;
            AUTUMN = 2;
            WINTER = 3;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    EnumType(
                        ["SPRING", "SUMMER", "AUTUMN", "WINTER"],
                        alias="_root.Season",
                    ),
                ],
                name="season",
                default=None,
            )
        ],
        alias="_root.WeatherReport",
    )


def test_to_recap_converter_oneof():
    protobuf_schema = """
        syntax = "proto3";
        message Contact {
            oneof contact_type {
                string email = 1;
                int64 phone_number = 2;
            }
        }
    """

    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StringType(variable=True, bytes_=2_147_483_648, name="email"),
                    IntType(bits=64, signed=True, name="phone_number"),
                ],
                name="contact_type",
                default=None,
            )
        ],
        alias="_root.Contact",
    )


def test_to_recap_converter_doubly_nested_message():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            message Middle {
                message Inner {
                    int32 innerValue = 1;
                }
                Inner middleValue = 2;
            }
            Middle outerValue = 3;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)

    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    StructType(
                                        [
                                            UnionType(
                                                [
                                                    NullType(),
                                                    IntType(bits=32, signed=True),
                                                ],
                                                name="innerValue",
                                                default=None,
                                            )
                                        ],
                                        alias="_root.Inner",
                                    ),
                                ],
                                name="middleValue",
                                default=None,
                            )
                        ],
                        alias="_root.Middle",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="_root.Outer",
    )


def test_to_recap_converter_timestamp():
    protobuf_schema = """
    syntax = "proto3";
    import "google/protobuf/timestamp.proto";
    message Test {
        google.protobuf.Timestamp timestamp = 1;
    }
    """
    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    IntType(
                        bits=64,
                        signed=True,
                        logical="build.recap.Timestamp",
                        unit="nanosecond",
                        timezone="UTC",
                    ),
                ],
                name="timestamp",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_to_recap_converter_duration():
    protobuf_schema = """
    syntax = "proto3";
    import "google/protobuf/duration.proto";
    message Test {
        google.protobuf.Duration duration = 1;
    }
    """
    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    IntType(
                        bits=64,
                        signed=True,
                        logical="build.recap.Duration",
                        unit="nanosecond",
                    ),
                ],
                name="duration",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_to_recap_converter_nullvalue():
    protobuf_schema = """
    syntax = "proto3";
    import "google/protobuf/struct.proto";
    message Test {
        google.protobuf.NullValue null_value = 1;
    }
    """
    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            NullType(
                name="null_value",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_to_recap_converter_self_reference():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            int32 value = 1;
            Outer next = 2;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)

    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    IntType(bits=32, signed=True),
                ],
                name="value",
                default=None,
            ),
            UnionType(
                [
                    NullType(),
                    ProxyType(
                        alias="_root.Outer",
                        registry=converter.registry,
                    ),
                ],
                name="next",
                default=None,
            ),
        ],
        alias="_root.Outer",
    )

    # Strip the alias since the resolved type won't have an alias.
    recap_schema.alias = None

    assert isinstance(recap_schema.fields[1], UnionType)
    assert isinstance(recap_schema.fields[1].types[1], ProxyType)
    assert recap_schema.fields[1].types[1].resolve() == recap_schema


def test_to_recap_converter_custom_namespace():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            message Inner {
                int32 innerValue = 1;
            }
            Inner outerValue = 2;
        }
    """

    converter = ProtobufConverter("foo.bar")
    recap_schema = converter.to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="foo.bar.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="foo.bar.Outer",
    )


def test_to_recap_converter_package():
    protobuf_schema = """
        syntax = "proto3";
        package foo.bar.baz;
        message Outer {
            message Inner {
                int32 innerValue = 1;
            }
            Inner outerValue = 2;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="foo.bar.baz.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="foo.bar.baz.Outer",
    )


def test_to_recap_converter_separate_files():
    protobuf_schema_1 = """
        syntax = "proto3";
        package foo.blah;
        message Inner {
            int32 innerValue = 1;
        }
    """
    protobuf_schema_2 = """
        syntax = "proto3";
        message Outer {
            foo.blah.Inner outerValue = 1;
        }
    """

    converter = ProtobufConverter()

    # Ignore return result. Just calling to load Inner into registry.
    converter.to_recap(protobuf_schema_1)

    recap_schema_2 = converter.to_recap(protobuf_schema_2)

    assert recap_schema_2 == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="foo.blah.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="_root.Outer",
    )


@pytest.mark.parametrize(
    "recap_type, expected_proto_type",
    [
        (NullType(name="some_field"), "google.protobuf.NullValue"),
        (BoolType(name="some_field"), "bool"),
        (IntType(signed=True, bits=32, name="some_field"), "int32"),
        (IntType(signed=True, bits=64, name="some_field"), "int64"),
        (IntType(signed=False, bits=32, name="some_field"), "uint32"),
        (IntType(signed=False, bits=64, name="some_field"), "uint64"),
        (FloatType(bits=32, name="some_field"), "float"),
        (FloatType(bits=64, name="some_field"), "double"),
        (StringType(bytes_=100, name="some_field"), "string"),
        (BytesType(bytes_=100, name="some_field"), "bytes"),
    ],
)
def test_from_recap(recap_type, expected_proto_type):
    converter = ProtobufConverter()
    struct_type = StructType(fields=[recap_type], alias="build.recap.MyStruct")
    result = converter.from_recap(struct_type)

    assert isinstance(result, ast.File)
    assert len(result.file_elements) == 2

    package = result.file_elements[0]
    assert isinstance(package, ast.Package)
    assert package.name == "build.recap"

    message = result.file_elements[1]
    assert isinstance(message, ast.Message)
    assert message.name == "MyStruct"
    assert len(message.elements) == 1

    field = message.elements[0]
    assert isinstance(field, ast.Field)
    assert field.name == "some_field"
    assert field.type == expected_proto_type
    assert field.number == 1


def test_from_recap_list_type():
    converter = ProtobufConverter()
    recap_type = ListType(
        values=IntType(signed=True, bits=32, name="some_int"),
        name="some_list",
    )
    struct_type = StructType(fields=[recap_type], alias="build.recap.MyStruct")

    result = converter.from_recap(struct_type)

    assert isinstance(result, ast.File)
    assert len(result.file_elements) == 2

    package = result.file_elements[0]
    assert isinstance(package, ast.Package)
    assert package.name == "build.recap"

    message = result.file_elements[1]
    assert isinstance(message, ast.Message)
    assert message.name == "MyStruct"
    assert len(message.elements) == 1

    field = message.elements[0]
    assert isinstance(field, ast.Field)
    assert field.name == "some_int"
    assert field.type == "int32"
    assert field.cardinality == ast.FieldCardinality.REPEATED
    assert field.number == 1


def test_from_recap_map_type():
    converter = ProtobufConverter()
    recap_type = MapType(
        keys=StringType(bytes_=100, name="map_key"),
        values=IntType(signed=True, bits=32, name="map_value"),
        name="some_map",
    )
    struct_type = StructType(fields=[recap_type], alias="build.recap.MyStruct")
    result = converter.from_recap(struct_type)

    assert isinstance(result, ast.File)
    assert len(result.file_elements) == 2

    package = result.file_elements[0]
    assert isinstance(package, ast.Package)
    assert package.name == "build.recap"

    message = result.file_elements[1]
    assert isinstance(message, ast.Message)
    assert message.name == "MyStruct"
    assert len(message.elements) == 1

    field = message.elements[0]
    assert isinstance(field, ast.MapField)
    assert field.name == "some_map"
    assert field.key_type == "string"
    assert field.value_type == "int32"
    assert field.number == 1


def test_from_recap_nullable():
    converter = ProtobufConverter()
    recap_type = UnionType(
        types=[NullType(), IntType(signed=True, bits=32, name="some_int")],
        name="some_union",
    )
    struct_type = StructType(fields=[recap_type], alias="build.recap.MyStruct")
    result = converter.from_recap(struct_type)

    assert isinstance(result, ast.File)
    assert len(result.file_elements) == 2

    package = result.file_elements[0]
    assert isinstance(package, ast.Package)
    assert package.name == "build.recap"

    message = result.file_elements[1]
    assert isinstance(message, ast.Message)
    assert message.name == "MyStruct"
    assert len(message.elements) == 1

    field = message.elements[0]
    assert isinstance(field, ast.Field)
    assert field.name == "some_int"
    assert field.type == "int32"
    assert field.number == 1


def test_from_recap_union():
    converter = ProtobufConverter()
    recap_type = UnionType(
        types=[
            IntType(signed=True, bits=32, name="some_int"),
            StringType(bytes_=100, name="some_string"),
        ],
        name="some_union",
    )
    struct_type = StructType(fields=[recap_type], alias="build.recap.MyStruct")
    result = converter.from_recap(struct_type)

    assert isinstance(result, ast.File)
    assert len(result.file_elements) == 2

    package = result.file_elements[0]
    assert isinstance(package, ast.Package)
    assert package.name == "build.recap"

    message = result.file_elements[1]
    assert isinstance(message, ast.Message)
    assert message.name == "MyStruct"
    assert len(message.elements) == 1

    oneof_field = message.elements[0]
    assert isinstance(oneof_field, ast.OneOf)
    assert oneof_field.name == "some_union"
    assert len(oneof_field.elements) == 2

    int_field = oneof_field.elements[0]
    assert isinstance(int_field, ast.Field)
    assert int_field.name == "some_int"
    assert int_field.type == "int32"
    assert int_field.number == 1

    string_field = oneof_field.elements[1]
    assert isinstance(string_field, ast.Field)
    assert string_field.name == "some_string"
    assert string_field.type == "string"
    assert string_field.number == 2


def test_from_recap_enum_type():
    converter = ProtobufConverter()
    recap_type = EnumType(
        name="some_enum",
        alias="build.recap.SomeEnum",
        symbols=[
            "ENUM_VALUE1",
            "ENUM_VALUE2",
        ],
    )
    struct_type = StructType(fields=[recap_type], alias="build.recap.MyStruct")
    result = converter.from_recap(struct_type)

    assert isinstance(result, ast.File)
    assert len(result.file_elements) == 3

    package = result.file_elements[0]
    assert isinstance(package, ast.Package)
    assert package.name == "build.recap"

    enum_type = result.file_elements[1]
    assert isinstance(enum_type, ast.Enum)
    assert enum_type.name == "SomeEnum"
    assert len(enum_type.elements) == 2
    assert isinstance(enum_type.elements[0], ast.EnumValue)
    assert enum_type.elements[0].name == "ENUM_VALUE1"
    assert isinstance(enum_type.elements[1], ast.EnumValue)
    assert enum_type.elements[1].name == "ENUM_VALUE2"

    message = result.file_elements[2]
    assert isinstance(message, ast.Message)
    assert message.name == "MyStruct"
    assert len(message.elements) == 1

    field = message.elements[0]
    assert isinstance(field, ast.Field)
    assert field.name == "some_enum"
    assert field.type == ".build.recap.SomeEnum"
    assert field.number == 1


def test_from_recap_complex_message():
    converter = ProtobufConverter()
    message_value_type = StructType(
        fields=[
            IntType(signed=True, bits=32, name="some_int"),
        ],
        # Nested type inside MyStruct
        alias="build.recap.MyStruct.ValueType",
    )
    recap_type = MapType(
        keys=StringType(bytes_=100),
        values=message_value_type,
        name="some_map",
    )
    struct_type = StructType(fields=[recap_type], alias="build.recap.MyStruct")
    result = converter.from_recap(struct_type)

    assert isinstance(result, ast.File)
    assert len(result.file_elements) == 2

    package = result.file_elements[0]
    assert isinstance(package, ast.Package)
    assert package.name == "build.recap"

    message = result.file_elements[1]
    assert isinstance(message, ast.Message)
    assert message.name == "MyStruct"
    assert len(message.elements) == 2

    map_field = message.elements[0]
    assert isinstance(map_field, ast.MapField)
    assert map_field.name == "some_map"
    assert map_field.key_type == "string"
    assert map_field.value_type == ".build.recap.MyStruct.ValueType"
    assert map_field.number == 1

    value_type = message.elements[1]
    assert isinstance(value_type, ast.Message)
    assert value_type.name == "ValueType"
    assert len(value_type.elements) == 1

    int_field = value_type.elements[0]
    assert isinstance(int_field, ast.Field)
    assert int_field.name == "some_int"
    assert int_field.type == "int32"
    assert int_field.number == 1


def test_from_recap_with_cyclic_reference():
    converter = ProtobufConverter()
    linked_list_node_type = StructType(
        fields=[
            IntType(bits=32, name="value"),
            ProxyType(
                alias="build.recap.LinkedListNode",
                registry=converter.registry,
                name="next",
            ),
        ],
        alias="build.recap.LinkedListNode",
    )
    result = converter.from_recap(linked_list_node_type)

    assert isinstance(result, ast.File)
    assert len(result.file_elements) == 2

    package = result.file_elements[0]
    assert isinstance(package, ast.Package)
    assert package.name == "build.recap"

    message = result.file_elements[1]
    assert isinstance(message, ast.Message)
    assert message.name == "LinkedListNode"
    assert len(message.elements) == 2

    value_field = message.elements[0]
    assert isinstance(value_field, ast.Field)
    assert value_field.name == "value"
    assert value_field.type == "int32"
    assert value_field.number == 1

    next_field = message.elements[1]
    assert isinstance(next_field, ast.Field)
    assert next_field.name == "next"
    assert next_field.type == ".build.recap.LinkedListNode"
    assert next_field.number == 2



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import Any, List, Protocol, Tuple

from recap.types import RecapType, StructType


class DbapiReader(ABC):
    def __init__(self, connection: Connection) -> None:
        self.connection = connection

    def to_recap(self, table: str, schema: str, catalog: str) -> StructType:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
                SELECT
                    *
                FROM information_schema.columns
                WHERE table_name = {self.param_style}
                    AND table_schema = {self.param_style}
                    AND table_catalog = {self.param_style}
                ORDER BY ordinal_position ASC
            """,
            (table, schema, catalog),
        )

        names = [name[0].upper() for name in cursor.description]
        fields = []
        for row in cursor.fetchall():
            column_props = dict(zip(names, row))
            base_type = self.get_recap_type(column_props)
            is_nullable = column_props["IS_NULLABLE"].upper() == "YES"

            if is_nullable:
                base_type = base_type.make_nullable()

            if column_props["COLUMN_DEFAULT"] is not None or is_nullable:
                # TODO Convert default value to the correct type, not just a string
                base_type.extra_attrs["default"] = column_props["COLUMN_DEFAULT"]

            base_type.extra_attrs["name"] = column_props["COLUMN_NAME"]

            fields.append(base_type)

        return StructType(fields=fields)

    @property
    def param_style(cls) -> str:
        return "%s"

    @abstractmethod
    def get_recap_type(self, column_props: dict[str, Any]) -> RecapType:
        ...

    def _get_time_unit(self, params: list[str] | None) -> str | None:
        match params:
            case [unit, _] if int(unit) == 0:
                return "second"
            case [unit, _] if int(unit) <= 3:
                return "millisecond"
            case [unit, _] if int(unit) <= 6:
                return "microsecond"

    def _parse_parameters(self, col_type: str) -> list[str] | None:
        """
        Parse types that have parameters.
        """
        match = re.search(r"\((.*?)\)", col_type)
        if match:
            return [p.strip() for p in match.group(1).split(",")]
        return None


class Connection(Protocol):
    def close(self) -> None:
        ...

    def commit(self) -> None:
        ...

    def rollback(self) -> None:
        ...

    def cursor(self) -> Cursor:
        ...


class Cursor(Protocol):
    def execute(self, query: str, parameters: Tuple = ()) -> None:
        ...

    def executemany(self, query: str, parameter_list: List[Tuple]) -> None:
        ...

    def fetchone(self) -> Tuple:
        ...

    def fetchall(self) -> List[Tuple]:
        ...

    def fetchmany(self, size: int) -> List[Tuple]:
        ...

    @property
    def description(self) -> List[Tuple]:
        ...



from __future__ import annotations

import json
import re
from typing import Any, Callable
from urllib.parse import urlparse

import httpx
from referencing import Registry, Resource, jsonschema, retrieval

from recap.types import (
    BUILTIN_ALIASES,
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    RecapType,
    RecapTypeRegistry,
    StringType,
    StructType,
    UnionType,
)

AliasStrategy = Callable[[str], str]


class JSONSchemaConverter:
    def __init__(self, json_registry: Registry | None = None):
        self.registry = RecapTypeRegistry()
        self.json_registry = json_registry or Registry(
            retrieve=JSONSchemaConverter.httpx_retrieve
        )

    def to_recap(
        self,
        json_schema_str: str,
        alias_strategy: AliasStrategy | None = None,
    ) -> StructType:
        alias_strategy = alias_strategy or JSONSchemaConverter.urn_to_alias
        json_schema = json.loads(json_schema_str)
        recap_schema = self._parse(json_schema, alias_strategy)
        if not isinstance(recap_schema, StructType):
            raise ValueError("JSON schema must be an object")
        return recap_schema

    def _parse(
        self,
        json_schema: dict,
        alias_strategy: AliasStrategy,
    ) -> RecapType:
        extra_attrs = {}
        if "description" in json_schema:
            extra_attrs["doc"] = json_schema["description"]
        if "default" in json_schema:
            extra_attrs["default"] = json_schema["default"]
        if "$id" in json_schema:
            resource = Resource.from_contents(json_schema, jsonschema.DRAFT202012)
            resource_id = resource.id()
            assert resource_id is not None, f"$id must be set for {json_schema}"
            self.json_registry = resource @ self.json_registry
            extra_attrs["alias"] = alias_strategy(resource_id)

        match json_schema:
            case {"type": "object", "properties": properties}:
                fields = []
                for name, prop in properties.items():
                    field = self._parse(prop, alias_strategy)
                    if name not in json_schema.get("required", []):
                        field = field.make_nullable()
                    field.extra_attrs["name"] = name
                    fields.append(field)
                return StructType(fields, **extra_attrs)
            case {"type": "array", "items": items}:
                values = self._parse(items, alias_strategy)
                return ListType(values, **extra_attrs)
            case {"type": "string", "format": "bytes"}:
                return BytesType(
                    bytes_=9_223_372_036_854_775_807,
                    **extra_attrs,
                )
            case {"type": "string", "format": "date"}:
                return StringType(
                    bytes_=9_223_372_036_854_775_807,
                    logical="org.iso.8601.Date",
                    **extra_attrs,
                )
            case {"type": "string", "format": "date-time"}:
                return StringType(
                    bytes_=9_223_372_036_854_775_807,
                    logical="org.iso.8601.DateTime",
                    **extra_attrs,
                )
            case {"type": "string", "format": "time"}:
                return StringType(
                    bytes_=9_223_372_036_854_775_807,
                    logical="org.iso.8601.Time",
                    **extra_attrs,
                )
            case {"type": "string"}:
                return StringType(bytes_=9_223_372_036_854_775_807, **extra_attrs)
            case {"type": "number"}:
                return FloatType(bits=64, **extra_attrs)
            case {"type": "integer"}:
                return IntType(bits=32, signed=True, **extra_attrs)
            case {"type": "boolean"}:
                return BoolType(**extra_attrs)
            case {"type": "null"}:
                return NullType(**extra_attrs)
            case {"$ref": str(urn)} | {"$dynamicRef": str(urn)}:
                try:
                    # If no KeyError is thrown here, we've already crawled this,
                    # so we can just return a ProxyType referencing the original.
                    resource = self.json_registry[urn]
                    assert (
                        resource.id() is not None
                    ), f"$id must be set for {resource.contents}"
                    alias = alias_strategy(str(resource.id()))
                    return ProxyType(alias, self.registry, **extra_attrs)
                except KeyError:
                    # We haven't crawled this type yet, so crawl and define the
                    # type as a normal Recap type with an alias set.
                    retrieved = self.json_registry.get_or_retrieve(urn)
                    resource = retrieved.value
                    self.json_registry = (
                        retrieved.registry @ self.json_registry
                    )  # pyright: ignore[reportGeneralTypeIssues]
                    return self._parse(resource.contents, alias_strategy)
            case _:
                raise ValueError(f"Unsupported JSON schema: {json_schema}")

    def from_recap(self, recap_type: RecapType) -> dict[str, Any]:
        """
        Convert a Recap type to a JSON schema. This method is recursive, so
        any nested types will be converted as well. The returned JSON schema
        will have a "$defs" key containing any aliased types.

        :param recap_type: The Recap type to convert.
        :return: A JSON schema.
        """

        type_dict = self._convert_field(recap_type)

        if defs := self._generate_alias_defs(recap_type):
            type_dict["$defs"] = defs

        return type_dict

    def _generate_alias_defs(self, recap_type: RecapType) -> dict[str, Any]:
        """
        Generate a dictionary of alias definitions for the given Recap type.
        Recursively generates definitions for any nested types, as well.
        This method is used to generate the "$defs" key in the top-level.

        :param recap_type: The Recap type to generate definitions for.
        :return: A dictionary of alias definitions.
        """

        defs = {}

        if (alias := recap_type.alias) and alias not in BUILTIN_ALIASES:
            # Don't use $ref for concrete types, since we're defining them,
            # not referencing them.
            defs[alias] = self._convert_field(recap_type, False)

        if isinstance(recap_type, StructType):
            for field in recap_type.fields:
                defs |= self._generate_alias_defs(field)
        elif isinstance(recap_type, ListType):
            defs |= self._generate_alias_defs(recap_type.values)
        elif isinstance(recap_type, UnionType):
            for value in recap_type.types:
                defs |= self._generate_alias_defs(value)
        elif isinstance(recap_type, MapType):
            defs |= self._generate_alias_defs(recap_type.values)

        return defs

    def _convert_field(
        self,
        field: RecapType,
        use_refs: bool = True,
    ) -> dict[str, Any]:
        """
        Convert a Recap type to a JSON schema type.

        :param field: The Recap type to convert.
        :param use_refs: If true, use $ref for concrete types with aliases.
        :return: A JSON schema type.
        """

        type_dict = {}

        if "default" in field.extra_attrs:
            type_dict |= {"default": field.extra_attrs["default"]}
        if doc := field.doc:
            type_dict |= {"description": doc}

        match field:
            case RecapType(alias=str(alias)) if use_refs and not isinstance(
                field,
                ProxyType,
            ):
                # This is a concrete type with an alias. Return the reference
                # since _generate_alias_defs will handle defining the type.
                type_dict = {"$ref": f"#/$defs/{alias}"}
            case UnionType(types=types):
                # If field is optional, just return the nested type.
                # from_recap will mark the field as required in the struct.
                if field.is_optional:
                    non_null_types = [t for t in types if not isinstance(t, NullType)]
                    assert len(non_null_types) == 1
                    type_dict |= self._convert_field(non_null_types[0])
                else:
                    type_dict |= {"oneOf": [self._convert_field(t) for t in types]}
            case StructType():
                type_dict |= {"type": "object", "properties": {}}
                for field in field.fields:
                    if field_name := field.extra_attrs.get("name"):
                        type_dict["properties"][field_name] = self._convert_field(field)

                        if not isinstance(field, UnionType) or not field.is_optional:
                            type_dict.setdefault("required", []).append(field_name)
                    else:
                        raise ValueError(f"Field {field} has no name")
            case ListType(values=values):
                type_dict |= {
                    "type": "array",
                    "items": self._convert_field(values),
                }
            case MapType(values=values):
                type_dict |= {
                    "type": "object",
                    "additionalProperties": self._convert_field(values),
                }
            case EnumType(symbols=symbols):
                type_dict |= {"type": "string", "enum": symbols}
            case BoolType():
                type_dict |= {"type": "boolean"}
            case IntType():
                type_dict |= {"type": "integer"}
            case FloatType():
                type_dict |= {"type": "number"}
            case StringType():
                type_dict |= {"type": "string"}
            case BytesType():
                type_dict |= {"type": "string", "format": "byte"}
            case NullType():
                type_dict |= {"type": "null"}
            case ProxyType(alias=alias) if alias in BUILTIN_ALIASES or len(
                field.extra_attrs
            ) > 0:
                # If there are attribute overrides, we can't use refs. JSON
                # schema doesn't support $def overrides. Also, if the alias
                # is a built-in type, convert to standard types instead of $defs
                return self._convert_field(field.resolve())
            case ProxyType(alias=str(alias)):
                return type_dict | {"$ref": f"#/$defs/{alias}"}
            case _:
                raise ValueError(f"Unsupported RecapType: {field.type_}")
        return type_dict

    @staticmethod
    def urn_to_alias(urn: str) -> str:
        """
        Convert a URN to a dotted alias name. The following rules are used:

        - The domain is reversed, e.g. "example.com" becomes "com.example"
        - The path is converted to lowercase and special characters are replaced
          with dots, e.g. "/foo/bar" becomes "foo.bar"
        - The domain and path are combined, e.g. "com.example/foo.bar"

        It's possible that this function will generate the same alias for
        different URNs. If this happens, you can provide a custom alias
        strategy to the `to_recap` method.

        :param urn: The URN to convert
        :return: The dotted alias
        """

        # Parse the URN/URL
        parsed = urlparse(urn)

        # Reverse domain
        domain_parts = parsed.netloc.split(".")
        domain_parts.reverse()
        domain = ".".join(domain_parts)

        # Replace special characters in the path, convert to lowercase
        path = parsed.path + parsed.fragment
        path = re.sub(r"[^a-zA-Z0-9]", ".", path)
        path = path.lower()

        # Combine domain and path
        dotted_type = domain
        if path and path != ".":
            dotted_type += "." + path.strip(".")

        return dotted_type

    @staticmethod
    @retrieval.to_cached_resource()
    def httpx_retrieve(uri: str):
        return httpx.get(uri).text



from proto_schema_parser.ast import (
    Enum,
    EnumElement,
    EnumValue,
    Field,
    FieldCardinality,
    File,
    MapField,
    Message,
    MessageElement,
    OneOf,
    Package,
)
from proto_schema_parser.parser import Parser

from recap.types import (
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    RecapType,
    RecapTypeRegistry,
    StringType,
    StructType,
    UnionType,
)

DEFAULT_NAMESPACE = "_root"
"""
Namespace to use when no namespace is specified in the schema.
"""


class ProtobufConverter:
    def __init__(self, namespace: str = DEFAULT_NAMESPACE) -> None:
        self.namespaces = [namespace]
        self.registry = RecapTypeRegistry()

    def to_recap(
        self,
        protobuf_schema_str: str,
    ) -> StructType:
        file = Parser().parse(protobuf_schema_str)
        root_message = self._parse(file)
        root_message = self._resolve_proxies(root_message)

        if not isinstance(root_message, StructType):
            raise ValueError("Protobuf schema must be a Message")

        return root_message

    def _resolve_proxies(
        self,
        recap_type: RecapType,
        resolved: set[str] | None = None,
        seen: set[str] | None = None,
    ) -> RecapType:
        resolved = resolved or set()
        seen = seen or set()

        # Detect circular references.
        if not isinstance(recap_type, ProxyType) and recap_type.alias is not None:
            seen.add(recap_type.alias)

        if (
            isinstance(recap_type, ProxyType)
            # Skip if already resolved.
            and recap_type.alias not in resolved
            # Skip if circular reference.
            and recap_type.alias not in seen
        ):
            extra_attrs = recap_type.extra_attrs
            alias = recap_type.alias
            recap_type = recap_type.resolve()
            recap_type.alias = alias
            recap_type.extra_attrs |= extra_attrs
            resolved.add(recap_type.alias)  # pyright: ignore[reportGeneralTypeIssues]

        if isinstance(recap_type, ListType):
            recap_type.values = self._resolve_proxies(recap_type.values, resolved, seen)
        elif isinstance(recap_type, MapType):
            recap_type.keys = self._resolve_proxies(recap_type.keys, resolved, seen)
            recap_type.values = self._resolve_proxies(recap_type.values, resolved, seen)
        elif isinstance(recap_type, UnionType):
            recap_type.types = [
                self._resolve_proxies(t, resolved, seen) for t in recap_type.types
            ]
        elif isinstance(recap_type, StructType):
            recap_type.fields = [
                self._resolve_proxies(f, resolved, seen) for f in recap_type.fields
            ]

        return recap_type

    def _parse(self, file: File) -> StructType:
        root_message = None

        # Find the namespace first, since we need it for message and enum aliases.
        packages = [p for p in file.file_elements if isinstance(p, Package)]

        if packages:
            self.namespaces.append(packages[0].name)

        # Now parse messages and enums.
        for file_element in file.file_elements:
            if isinstance(file_element, Message):
                struct_type = self._parse_message(file_element)
                root_message = root_message or struct_type
            elif isinstance(file_element, Enum):
                self._parse_enum(file_element)

        if root_message is None:
            raise ValueError("Protobuf schema must contain at least one Message")

        self.namespaces.pop()

        return root_message

    def _parse_message(self, message: Message) -> StructType:
        fields = []

        for element in message.elements:
            if isinstance(element, Field):
                recap_type = self._parse_field(element)
                fields.append(recap_type)
            elif isinstance(element, MapField):
                recap_type = self._parse_map_field(element)
                fields.append(recap_type)
            elif isinstance(element, OneOf):
                recap_type = self._parse_oneof(element)
                fields.append(recap_type)
            elif isinstance(element, Message):
                self._parse_message(element)
            elif isinstance(element, Enum):
                self._parse_enum(element)

        struct_type = StructType(
            fields=fields,
            alias=f"{self.namespaces[-1]}.{message.name}",
        )

        self.registry.register_alias(struct_type)

        return struct_type

    def _parse_field(self, field: Field) -> RecapType:
        recap_type = self._protobuf_type_to_recap_type(field.type)

        if field.cardinality == FieldCardinality.REPEATED:
            recap_type = ListType(values=recap_type)

        if field.cardinality != FieldCardinality.REQUIRED:
            if not isinstance(recap_type, NullType):
                recap_type = UnionType(types=[NullType(), recap_type], default=None)
            recap_type.extra_attrs["default"] = None

        recap_type.extra_attrs["name"] = field.name

        return recap_type

    def _parse_map_field(self, field: MapField) -> RecapType:
        key_type = self._protobuf_type_to_recap_type(field.key_type)

        # Proto map keys are always strings or integers.
        assert isinstance(key_type, IntType) or isinstance(key_type, StringType)

        value_type = self._protobuf_type_to_recap_type(field.value_type)
        map_type = UnionType(
            types=[
                NullType(),
                MapType(keys=key_type, values=value_type),
            ],
            default=None,
        )

        if field.name is not None:
            map_type.extra_attrs["name"] = field.name

        return map_type

    def _protobuf_type_to_recap_type(self, protobuf_type: str) -> RecapType:
        # NOTE: Protobuf doesn't support type aliases, so we don't need to
        # register aliases here (or in oneof, map, etc).
        match protobuf_type:
            case "string":
                return StringType(bytes_=2_147_483_648, variable=True)
            case "bool":
                return BoolType()
            case "bytes":
                return BytesType(bytes_=2_147_483_648, variable=True)
            case "int32":
                return IntType(bits=32)
            case "int64":
                return IntType(bits=64)
            case "uint32":
                return IntType(bits=32, signed=False)
            case "uint64":
                return IntType(bits=64, signed=False)
            case "sint32":
                return IntType(bits=32)
            case "sint64":
                return IntType(bits=64)
            case "float":
                return FloatType(bits=32)
            case "double":
                return FloatType(bits=64)
            case "fixed32":
                return IntType(bits=32, signed=False)
            case "fixed64":
                return IntType(bits=64, signed=False)
            case "sfixed32":
                return IntType(bits=32)
            case "sfixed64":
                return IntType(bits=64)
            # Handle some of Google's well-known types.
            # Technically, we should honor `import google/protobuf/*.proto` and
            # make sure the protos match what's expected, but assuming they match
            # is fine for now.
            case "google.protobuf.Timestamp":
                # https://protobuf.dev/reference/protobuf/google.protobuf/#timestamp
                # Note: Some protobuf values might overrun this since they support
                # from 0001-01-01T00:00:00Z to 9999-12-31T23:59:59.999999999Z.
                return IntType(
                    logical="build.recap.Timestamp",
                    bits=64,
                    signed=True,
                    unit="nanosecond",
                    timezone="UTC",
                )
            case "google.protobuf.Duration":
                # https://protobuf.dev/reference/protobuf/google.protobuf/#duration
                # Note: Some protobuf values might overrun this since they support
                # int64 seconds and int32 nanoseconds.
                return IntType(
                    bits=64,
                    logical="build.recap.Duration",
                    unit="nanosecond",
                )
            case "google.protobuf.NullValue":
                return NullType()
            case _:
                # Create a ProxyType and hope we have a type alias for this type.
                alias = (
                    protobuf_type
                    if "." in protobuf_type
                    else f"{self.namespaces[-1]}.{protobuf_type}"
                )
                return ProxyType(alias, self.registry)

    def _parse_oneof(self, oneof: OneOf) -> RecapType:
        types: list[RecapType] = [NullType()]
        for element in oneof.elements:
            if isinstance(element, Field):
                # !! HACK !!
                # Force oneof subfield to be required since we're unioning with null above.
                # Otherwise we end up with [NullType, [NullType, RecapType], ...].
                # This is safe to force because oneof subfields can't be repeated.
                original_cardinality = element.cardinality
                element.cardinality = FieldCardinality.REQUIRED
                recap_type = self._parse_field(element)
                element.cardinality = original_cardinality
                types.append(recap_type)
        return UnionType(types=types, name=oneof.name, default=None)

    def _parse_enum(self, enum: Enum) -> RecapType:
        symbols = []
        for element in enum.elements:
            if isinstance(element, EnumValue):
                symbols.append(element.name)
        enum_type = EnumType(
            symbols=symbols,
            alias=f"{self.namespaces[-1]}.{enum.name}",
        )
        self.registry.register_alias(enum_type)
        return enum_type

    def from_recap(self, struct_type: StructType) -> File:
        """Convert a Recap type to a Protobuf type."""
        assert struct_type.alias is not None, "Struct must have an alias."
        package_to_types: dict[str, list[Message | Enum]] = {}
        self._from_recap_gather_types(struct_type, package_to_types)
        # Lexicographical sort always puts the root package first.
        package = sorted(package_to_types.keys())[0]
        # Verify all subpackages are under the root package.
        for subpackage in package_to_types.keys():
            assert subpackage.startswith(package), (
                "Proto files must contain messages in a single root package space, "
                f"but found root={package} and subpackage={subpackage}."
            )
        return File(
            file_elements=[Package(name=package)]
            + self._from_recap_construct_elements(package, package_to_types)
        )

    def _from_recap_gather_types(
        self,
        recap_type: RecapType,
        package_to_types: dict[str, list[Message | Enum]],
    ) -> None:
        """
        Find all Message and Enum types that need to be declared in the proto
        file. Keep track of their package so we can construct the proto file
        with the correct nesting.

        :param recap_type: The Recap type to convert.
        :param package_to_types: A mapping of package to types that have been
            discovered.
        """

        match recap_type:
            case StructType(fields=fields):
                assert recap_type.alias is not None, "Struct must have an alias."
                package, message_name = recap_type.alias.rsplit(".", 1)
                field_number = 1
                message_elements: list[MessageElement] = []

                for field in recap_type.fields:
                    message_elements.append(
                        self._recap_to_field(field, field_number, package)
                    )
                    field_number += 1
                    self._from_recap_gather_types(field, package_to_types)

                package_to_types.setdefault(package, []).append(
                    Message(
                        message_name,
                        message_elements,
                    )
                )
            case UnionType(types=types):
                for type_ in types:
                    self._from_recap_gather_types(type_, package_to_types)
            case MapType(keys=keys, values=values):
                self._from_recap_gather_types(keys, package_to_types)
                self._from_recap_gather_types(values, package_to_types)
            case ListType(values=values):
                self._from_recap_gather_types(values, package_to_types)
            case EnumType(symbols=symbols):
                assert recap_type.alias is not None, "Enum must have an alias."
                enum_package, enum_name = recap_type.alias.rsplit(".", 1)
                elements: list[EnumElement] = [
                    EnumValue(symbol, idx)
                    for symbol, idx in zip(
                        recap_type.symbols, range(len(recap_type.symbols))
                    )
                ]
                enum = Enum(enum_name, elements)
                package_to_types.setdefault(enum_package, []).append(enum)

    def _from_recap_construct_elements(
        self,
        package: str,
        package_to_types: dict[str, list[Message | Enum]],
    ) -> list[Message | Enum]:
        """
        Construct the proto file elements for a given package. This method
        mutates the existing Message/Enum structures to contain nested
        elements.

        :param package: The package to construct elements for.
        :param package_to_types: A mapping of package to types that have been
            discovered.
        :return: The proto FileElements (or MessageElements) for the package.
        """

        package_types = package_to_types.get(package, [])
        for package_type in package_types:
            if isinstance(package_type, Message):
                subelements = self._from_recap_construct_elements(
                    f"{package}.{package_type.name}",
                    package_to_types,
                )
                package_type.elements.extend(subelements)
        return package_types

    def _recap_to_field(
        self,
        recap_type: RecapType,
        field_number: int,
        package: str,
    ) -> Field | MapField | OneOf:
        """
        Convert a Recap type to a Protobuf field. This method is recursive and
        will call itself for nested types.

        :param recap_type: The Recap type to convert.
        :param field_number: The field number to assign to the field.
        :param package: The package of the field.
        :return: The Protobuf field.
        """

        field_name = recap_type.extra_attrs.get("name", f"field_{field_number}")
        match recap_type:
            case NullType():
                return Field(field_name, field_number, "google.protobuf.NullValue")
            case BoolType():
                return Field(field_name, field_number, "bool")
            case IntType(signed=bool(signed), bits=int(bits)) if signed and bits <= 32:
                return Field(field_name, field_number, "int32")
            case IntType(signed=bool(signed), bits=int(bits)) if signed and bits <= 64:
                return Field(field_name, field_number, "int64")
            case IntType(
                signed=bool(signed), bits=int(bits)
            ) if not signed and bits <= 32:
                return Field(field_name, field_number, "uint32")
            case IntType(
                signed=bool(signed), bits=int(bits)
            ) if not signed and bits <= 64:
                return Field(field_name, field_number, "uint64")
            case IntType():
                raise ValueError(
                    "Invalid bit size for IntType. Protobuf supports only 32 "
                    "or 64 bits for integer types."
                )
            case FloatType(bits=int(bits)) if bits <= 32:
                return Field(field_name, field_number, "float")
            case FloatType(bits=int(bits)) if bits <= 64:
                return Field(field_name, field_number, "double")
            case FloatType():
                raise ValueError(
                    "Invalid bit size for FloatType. Protobuf supports only 32 "
                    "or 64 bits for floating-point types."
                )
            case StringType(bytes_=int(bytes_)) if bytes_ <= 2_147_483_647:
                return Field(field_name, field_number, "string")
            case BytesType(bytes_=int(bytes_)) if bytes_ <= 2_147_483_647:
                return Field(field_name, field_number, "bytes")
            case ListType(values=values):
                nested_field = self._recap_to_field(values, field_number, package)
                assert isinstance(
                    nested_field,
                    Field,
                ), "Maps and unions can't be repeated (lists)."
                nested_field.cardinality = FieldCardinality.REPEATED
                return nested_field
            case MapType(keys=keys, values=values):
                # 0 field_number because we ignore it. We just need types.
                key_field = self._recap_to_field(keys, 0, package)
                value_field = self._recap_to_field(values, 0, package)
                assert isinstance(
                    key_field, Field
                ), "Key must be integral or string type."
                assert isinstance(value_field, Field), "Value must not be a map type."
                return MapField(
                    field_name,
                    field_number,
                    key_field.type,
                    value_field.type,
                )
            case UnionType(types=types):
                non_null_types = [t for t in types if not isinstance(t, NullType)]
                if len(non_null_types) == 1:
                    # One non-null type means this is a nullable field.
                    field = self._recap_to_field(
                        non_null_types[0],
                        field_number,
                        package,
                    )
                    assert isinstance(field, Field), "Maps can't be nullable."
                    field.cardinality = FieldCardinality.OPTIONAL
                    return field
                else:
                    # Multiple non-null types means this is a true union of multiple types.
                    oneof_field_num = field_number
                    oneof_fields = []
                    # OneOf fields are always optional, so no need to include NullType.
                    for non_null_type in non_null_types:
                        field = self._recap_to_field(
                            non_null_type,
                            oneof_field_num,
                            package,
                        )
                        oneof_fields.append(field)
                        oneof_field_num += 1
                    return OneOf(field_name, oneof_fields)
            case (
                StructType(alias=alias) | EnumType(alias=alias) | ProxyType(alias=alias)
            ):
                assert alias is not None, "Expected alias to be set."
                # Always start from the root since Recap namespaces are always root-based.
                return Field(field_name, field_number, "." + alias)
            case _:
                raise ValueError(f"Invalid RecapType: {recap_type}")



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from confluent_kafka.schema_registry import SchemaRegistryClient

from recap.converters.avro import AvroConverter
from recap.converters.json_schema import JSONSchemaConverter
from recap.converters.protobuf import ProtobufConverter
from recap.types import StructType


class ConfluentRegistryReader:
    def __init__(self, registry: SchemaRegistryClient):
        self.registry = registry

    @staticmethod
    @contextmanager
    def create(**kwargs) -> Generator[ConfluentRegistryReader, None, None]:
        with SchemaRegistryClient(**kwargs) as registry:
            yield ConfluentRegistryReader(registry)

    def to_recap(self, topic: str) -> StructType:
        subject = f"{topic}-value"
        registered_schema = self.registry.get_latest_version(subject)
        schema_str = registered_schema.schema.schema_str
        match registered_schema.schema.schema_type:
            case "AVRO":
                return AvroConverter().to_recap(schema_str)
            case "JSON":
                return JSONSchemaConverter().to_recap(schema_str)
            case "PROTOBUF":
                return ProtobufConverter().to_recap(schema_str)
            case _:
                raise ValueError(
                    f"Unsupported schema type {registered_schema.schema.schema_type}"
                )



from __future__ import annotations

from contextlib import contextmanager
from math import ceil
from typing import Any, Generator

from recap.readers.dbapi import DbapiReader
from recap.types import BoolType, BytesType, FloatType, IntType, RecapType, StringType

MAX_FIELD_SIZE = 1073741824


class PostgresqlReader(DbapiReader):
    @staticmethod
    @contextmanager
    def create(**kwargs) -> Generator[PostgresqlReader, None, None]:
        import psycopg2

        with psycopg2.connect(**kwargs) as client:
            yield PostgresqlReader(client)

    def get_recap_type(self, column_props: dict[str, Any]) -> RecapType:
        data_type = column_props["DATA_TYPE"].lower()
        octet_length = column_props["CHARACTER_OCTET_LENGTH"]
        max_length = column_props["CHARACTER_MAXIMUM_LENGTH"]

        if data_type in ["bigint", "int8", "bigserial", "serial8"]:
            base_type = IntType(bits=64, signed=True)
        elif data_type in ["integer", "int", "int4", "serial", "serial4"]:
            base_type = IntType(bits=32, signed=True)
        elif data_type in ["smallint", "smallserial", "serial2"]:
            base_type = IntType(bits=16, signed=True)
        elif data_type in ["double precision", "float8"]:
            base_type = FloatType(bits=64)
        elif data_type in ["real", "float4"]:
            base_type = FloatType(bits=32)
        elif data_type == "boolean":
            base_type = BoolType()
        elif (
            data_type in ["text", "json", "jsonb"]
            or data_type.startswith("character varying")
            or data_type.startswith("varchar")
        ):
            base_type = StringType(bytes_=octet_length, variable=True)
        elif data_type.startswith("char"):
            base_type = StringType(bytes_=octet_length, variable=False)
        elif data_type == "uuid":
            base_type = StringType(
                logical="build.recap.UUID",
                bytes_=36,
                variable=False,
            )
        elif data_type == "bytea" or data_type.startswith("bit varying"):
            base_type = BytesType(bytes_=MAX_FIELD_SIZE, variable=True)
        elif data_type.startswith("bit"):
            byte_length = ceil(max_length / 8)
            base_type = BytesType(bytes_=byte_length, variable=False)
        elif data_type.startswith("timestamp"):
            dt_precision = column_props["DATETIME_PRECISION"]
            unit = self._get_time_unit([dt_precision]) or "microsecond"
            base_type = IntType(
                bits=64,
                logical="build.recap.Timestamp",
                unit=unit,
            )
        elif data_type.startswith("decimal") or data_type.startswith("numeric"):
            base_type = BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=column_props["NUMERIC_PRECISION"],
                scale=column_props["NUMERIC_SCALE"],
            )
        else:
            raise ValueError(f"Unknown data type: {data_type}")

        return base_type



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Generator

from pymetastore.htypes import (
    HCharType,
    HDecimalType,
    HListType,
    HMapType,
    HPrimitiveType,
    HStructType,
    HType,
    HUnionType,
    HVarcharType,
    PrimitiveCategory,
)
from pymetastore.metastore import HMS, HColumn, HTable
from pymetastore.stats import (
    BinaryTypeStats,
    BooleanTypeStats,
    DateTypeStats,
    DecimalTypeStats,
    DoubleTypeStats,
    LongTypeStats,
    StringTypeStats,
    TypeStats,
)

from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    RecapType,
    StringType,
    StructType,
    UnionType,
)


class HiveMetastoreReader:
    def __init__(self, client: HMS):
        self.client = client

    @staticmethod
    @contextmanager
    def create(**kwargs) -> Generator[HiveMetastoreReader, None, None]:
        with HMS.create(**kwargs) as client:
            yield HiveMetastoreReader(client)

    def to_recap(
        self,
        database_name: str,
        table_name: str,
        include_stats: bool = False,
    ) -> StructType:
        table = self.client.get_table(database_name, table_name)
        fields: list[RecapType] = []
        for col in table.columns:
            recap_type = self._convert(col)
            fields.append(recap_type)
        if include_stats:
            self._load_stats(table, fields)
        return StructType(fields=fields, name=table_name)

    def _convert(self, htype: HType | HColumn) -> RecapType:
        recap_type: RecapType | None = None
        extra_attrs = {}
        if isinstance(htype, HColumn):
            extra_attrs["name"] = htype.name
            if htype.comment:
                extra_attrs["comment"] = htype.comment
            htype = htype.type
        match htype:
            case HPrimitiveType(primitive_type=PrimitiveCategory.BOOLEAN):
                recap_type = BoolType()
            case HPrimitiveType(primitive_type=PrimitiveCategory.BYTE):
                recap_type = IntType(bits=8)
            case HPrimitiveType(primitive_type=PrimitiveCategory.SHORT):
                recap_type = IntType(bits=16)
            case HPrimitiveType(primitive_type=PrimitiveCategory.INT):
                recap_type = IntType(bits=32)
            case HPrimitiveType(primitive_type=PrimitiveCategory.LONG):
                recap_type = IntType(bits=64)
            case HPrimitiveType(primitive_type=PrimitiveCategory.FLOAT):
                recap_type = FloatType(bits=32)
            case HPrimitiveType(primitive_type=PrimitiveCategory.DOUBLE):
                recap_type = FloatType(bits=64)
            case HPrimitiveType(primitive_type=PrimitiveCategory.VOID):
                recap_type = NullType(**extra_attrs)
            case HPrimitiveType(primitive_type=PrimitiveCategory.STRING):
                # TODO: Should handle multi-byte encodings
                # Using 2^63-1 as the max length because Hive has no defined max.
                recap_type = StringType(bytes_=9_223_372_036_854_775_807)
            case HPrimitiveType(primitive_type=PrimitiveCategory.BINARY):
                recap_type = BytesType(bytes_=2_147_483_647)
            case HDecimalType(precision=int(precision), scale=int(scale)):
                # TODO: Decimals can be > 38 digits, but need to calculate bytes_.
                if precision > 38:
                    raise ValueError(f"Max decimal precision is 38, got {precision}")
                recap_type = BytesType(
                    logical="build.recap.Decimal",
                    bytes_=16,
                    variable=False,
                    precision=precision,
                    scale=scale,
                )
            case HVarcharType(length=int(length)):
                # TODO: Should handle multi-byte encodings
                recap_type = StringType(bytes_=length)
            case HCharType(length=int(length)):
                # TODO: Should handle multi-byte encodings
                recap_type = StringType(bytes_=length, variable=False)
            case HPrimitiveType(primitive_type=PrimitiveCategory.DATE):
                recap_type = IntType(
                    logical="build.recap.Date",
                    bits=32,
                    signed=True,
                    unit="day",
                    **extra_attrs,
                )
            case HPrimitiveType(primitive_type=PrimitiveCategory.TIMESTAMP):
                # Hive stores timestamps as strings, so 64 bits isn't enough to store
                # all possible values. C'est la vie.
                recap_type = IntType(
                    logical="build.recap.Timestamp",
                    bits=64,
                    signed=True,
                    unit="nanosecond",
                    timezone="UTC",
                    **extra_attrs,
                )
            case HPrimitiveType(primitive_type=PrimitiveCategory.TIMESTAMPLOCALTZ):
                # Hive stores timestamps as strings, so 64 bits isn't enough to store
                # all possible values. C'est la vie.
                recap_type = IntType(
                    logical="build.recap.Timestamp",
                    bits=64,
                    signed=True,
                    unit="nanosecond",
                    # Follow Apache Arrow's Schema.fbs style for local time.
                    # Denote LOCALTZ as a timezone field set to None.
                    timezone=None,
                    **extra_attrs,
                )
            case HPrimitiveType(primitive_type=PrimitiveCategory.INTERVAL_YEAR_MONTH):
                recap_type = BytesType(
                    logical="build.recap.Interval",
                    bytes_=12,
                    signed=True,
                    unit="month",
                    **extra_attrs,
                )
            case HPrimitiveType(primitive_type=PrimitiveCategory.INTERVAL_DAY_TIME):
                recap_type = BytesType(
                    logical="build.recap.Interval",
                    bytes_=12,
                    signed=True,
                    unit="second",
                    **extra_attrs,
                )
            case HMapType(key_type=key_type, value_type=value_type):
                recap_type = MapType(
                    keys=self._convert(key_type),
                    values=self._convert(value_type),
                )
            case HListType(element_type=element_type):
                recap_type = ListType(values=self._convert(element_type))
            case HUnionType(types=types):
                # Hive columns are always nullable and default is always null.
                types = [NullType()] + [self._convert(t) for t in htype.types]
                recap_type = UnionType(types=types, default=None, **extra_attrs)
                recap_type = self._flatten_unions(recap_type)
            case HStructType(names=names, types=types):
                recap_type = StructType(
                    fields=[self._convert(HColumn(n, t)) for n, t in zip(names, types)]
                )
            case _:
                raise ValueError(f"Unsupported type: {htype}")

        if not isinstance(recap_type, UnionType) and not isinstance(
            recap_type, NullType
        ):
            # Hive columns are always nullable and default is always null.
            recap_type = UnionType(
                types=[NullType(), recap_type],
                default=None,
                **extra_attrs,
            )

        return recap_type

    def _flatten_unions(self, recap_type: RecapType) -> RecapType:
        """
        Flattens nested unions into a single union. Useful for unwinding a
        HUnionType since we're wrapping all fields in _convert() in a union.
        Without this, we'd get [NullType, [NullType, IntType], [NullType,
        StrType], ...]. Now we get [NullType, IntType, StrType].
        """

        if isinstance(recap_type, UnionType):
            # Start with empty set of flattened types
            flattened_types = []

            for union_subtype in recap_type.types:
                # We're not expecting string type aliases here.
                assert isinstance(union_subtype, RecapType)

                # Recursively flatten unions within unions
                flattened = self._flatten_unions(union_subtype)

                if isinstance(flattened, UnionType):
                    # Append types from inner union to the flattened types
                    flattened_types.extend(
                        [t for t in flattened.types if t not in flattened_types]
                    )
                elif flattened not in flattened_types:
                    # Append non-union types directly
                    flattened_types.append(flattened)

            return UnionType(types=flattened_types, **recap_type.extra_attrs)
        elif isinstance(recap_type, StructType):
            return StructType(
                fields=[self._flatten_unions(field) for field in recap_type.fields],
                **recap_type.extra_attrs,
            )
        elif isinstance(recap_type, ListType):
            return ListType(
                values=self._flatten_unions(recap_type.values),
                extra_attrs=recap_type.extra_attrs,
            )
        elif isinstance(recap_type, MapType):
            return MapType(
                keys=self._flatten_unions(recap_type.keys),
                values=self._flatten_unions(recap_type.values),
                extra_attrs=recap_type.extra_attrs,
            )
        else:
            # No nested union type to flatten
            return recap_type

    def _load_stats(self, table: HTable, fields: list[RecapType]) -> None:
        fields_dict = {f.extra_attrs["name"]: f for f in fields}
        table_stats = self.client.get_table_stats(table, [])
        table_stats = [cs for cs in table_stats if cs.stats is not None]
        for col_stats in table_stats:
            recap_type = fields_dict[col_stats.columnName]
            match col_stats.stats:
                case LongTypeStats() | DoubleTypeStats():
                    recap_type.extra_attrs |= {
                        "low": col_stats.stats.lowValue,
                        "high": col_stats.stats.highValue,
                        "cardinality": col_stats.stats.cardinality,
                    }
                case DateTypeStats():
                    stats: dict[str, Any] = {
                        "cardinality": col_stats.stats.cardinality,
                    }
                    if col_stats.stats.lowValue is not None:
                        stats["low"] = col_stats.stats.lowValue.daysSinceEpoch
                    if col_stats.stats.highValue is not None:
                        stats["high"] = col_stats.stats.highValue.daysSinceEpoch
                    recap_type.extra_attrs |= stats
                case StringTypeStats():
                    recap_type.extra_attrs |= {
                        "average_length": col_stats.stats.avgColLen,
                        "max_length": col_stats.stats.maxColLen,
                        "cardinality": col_stats.stats.cardinality,
                    }
                case BooleanTypeStats():
                    recap_type.extra_attrs |= {
                        "true_count": col_stats.stats.numTrues,
                        "false_count": col_stats.stats.numFalses,
                    }
                case BinaryTypeStats():
                    recap_type.extra_attrs |= {
                        "average_length": col_stats.stats.avgColLen,
                        "max_length": col_stats.stats.maxColLen,
                    }
                case DecimalTypeStats():
                    stats: dict[str, Any] = {
                        "cardinality": col_stats.stats.cardinality,
                    }
                    if col_stats.stats.lowValue is not None:
                        stats["low"] = col_stats.stats.lowValue
                    if col_stats.stats.highValue is not None:
                        stats["high"] = col_stats.stats.highValue
                    recap_type.extra_attrs |= stats

            if isinstance(col_stats.stats, TypeStats):
                recap_type.extra_attrs["null_count"] = col_stats.stats.numNulls



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from google.cloud import bigquery

from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    StringType,
    StructType,
)


class BigQueryReader:
    def __init__(self, client: bigquery.Client):
        self.client = client

    @staticmethod
    @contextmanager
    def create(**kwargs) -> Generator[BigQueryReader, None, None]:
        with bigquery.Client(**kwargs) as client:
            yield BigQueryReader(client)

    def to_recap(self, dataset: str, table: str) -> StructType:
        table_ref = self.client.dataset(dataset).table(table)
        table_obj = self.client.get_table(table_ref)
        return self._parse_fields(table_obj.schema)

    def _parse_fields(self, fields: list[bigquery.SchemaField]) -> StructType:
        recap_fields = []
        for field in fields:
            match field.field_type:
                case "STRING" | "JSON":
                    # I'm having a hard time finding the max string length in BQ. This link:
                    # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types
                    # Says "2 logical bytes + the UTF-8 encoded string size", so I'm assuming
                    # the 2 logical bytes are a uint16 length header, which is 65_536.
                    field_type = StringType(bytes_=field.max_length or 65_536)
                case "BYTES":
                    field_type = BytesType(bytes_=field.max_length or 65_536)
                case "INT64" | "INTEGER" | "INT" | "SMALLINT" | "TINYINT" | "BYTEINT":
                    field_type = IntType(bits=64)
                case "FLOAT" | "FLOAT64":
                    field_type = FloatType(bits=64)
                case "BOOLEAN":
                    field_type = BoolType()
                case "TIMESTAMP" | "DATETIME":
                    field_type = IntType(
                        logical="build.recap.Timestamp",
                        bits=64,
                        unit="microsecond",
                    )
                case "TIME":
                    field_type = IntType(
                        logical="build.recap.Time",
                        bits=32,
                        unit="microsecond",
                    )
                case "DATE":
                    field_type = IntType(
                        logical="build.recap.Date",
                        bits=32,
                        unit="day",
                    )
                case "RECORD" | "STRUCT":
                    field_type = self._parse_fields(list(field.fields))
                case "NUMERIC" | "DECIMAL":
                    field_type = BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=field.precision or 38,
                        scale=field.scale or 0,
                    )
                case "BIGNUMERIC" | "BIGDECIMAL":
                    field_type = BytesType(
                        logical="build.recap.Decimal",
                        bytes_=32,
                        variable=False,
                        precision=field.precision or 76,
                        scale=field.scale or 0,
                    )
                case _:
                    raise ValueError(f"Unrecognized field type: {field.field_type}")

            if field.mode == "REPEATED":
                field_type = ListType(field_type)
            if field.mode == "NULLABLE":
                field_type = field_type.make_nullable()
            if name := field.name:
                field_type.extra_attrs["name"] = name
            if description := field.description:
                field_type.doc = description
            if default := field.default_value_expression:
                field_type.extra_attrs["default"] = default

            recap_fields.append(field_type)

        return StructType(recap_fields)



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Generator

from recap.readers.dbapi import DbapiReader
from recap.types import BytesType, FloatType, IntType, RecapType, StringType


class MysqlReader(DbapiReader):
    @staticmethod
    @contextmanager
    def create(**kwargs) -> Generator[MysqlReader, None, None]:
        import mysql.connector

        with mysql.connector.connect(**kwargs) as client:
            yield MysqlReader(client)  # pyright: ignore[reportGeneralTypeIssues]

    def get_recap_type(self, column_props: dict[str, Any]) -> RecapType:
        data_type = column_props["DATA_TYPE"].lower()
        octet_length = column_props["CHARACTER_OCTET_LENGTH"]
        precision = column_props["NUMERIC_PRECISION"]
        scale = column_props["NUMERIC_SCALE"]

        if data_type == "bigint":
            # https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
            base_type = IntType(bits=64, signed=True)
        elif data_type in ["int", "integer"]:
            # https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
            base_type = IntType(bits=32, signed=True)
        elif data_type == "mediumint":
            # https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
            base_type = IntType(bits=24, signed=True)
        elif data_type == "smallint":
            # https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
            base_type = IntType(bits=16, signed=True)
        elif data_type == "tinyint":
            # https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
            base_type = IntType(bits=8, signed=True)
        elif data_type == "double" or (data_type == "float" and precision > 23):
            # https://dev.mysql.com/doc/refman/8.0/en/floating-point-types.html
            base_type = FloatType(bits=64)
        elif data_type == "float" and precision <= 23:
            # https://dev.mysql.com/doc/refman/8.0/en/floating-point-types.html
            base_type = FloatType(bits=32)
        elif data_type in [
            "text",
            "json",
            "mediumtext",
            "longtext",
            "tinytext",
        ] or data_type.startswith("varchar"):
            # This isn't exactly correct for JSON. MySQL uses a binary encoding
            # to store the structure.
            base_type = StringType(bytes_=octet_length, variable=True)
        elif data_type.startswith("char") or data_type in ["enum", "set"]:
            base_type = StringType(bytes_=octet_length, variable=False)
        elif data_type in [
            "blob",
            "mediumblob",
            "longblob",
            "tinyblob",
        ] or data_type.startswith("varbinary"):
            base_type = BytesType(bytes_=octet_length, variable=True)
        elif data_type.startswith("binary"):
            base_type = BytesType(bytes_=octet_length, variable=False)
        elif data_type.startswith("bit"):
            base_type = BytesType(bytes_=8, variable=False)
        elif data_type.startswith("timestamp") or data_type.startswith("datetime"):
            dt_precision = column_props["DATETIME_PRECISION"]
            unit = self._get_time_unit([dt_precision]) or "microsecond"
            base_type = IntType(
                bits=64,
                logical="build.recap.Timestamp",
                unit=unit,
            )
        elif data_type in ["dec", "decimal", "numeric"]:
            base_type = BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=precision,
                scale=scale,
            )
        elif data_type == "year":
            base_type = IntType(
                bits=16,
                signed=False,
            )  # Years are typically 2-byte values
        else:
            raise ValueError(f"Unknown data type: {data_type}")

        return base_type



from unittest.mock import MagicMock

from confluent_kafka import schema_registry

from recap.readers.confluent_registry import ConfluentRegistryReader
from recap.types import IntType, StringType, StructType, UnionType


def test_struct_avro():
    mock_schema_registry_client = MagicMock(spec=schema_registry.SchemaRegistryClient)

    class MockSchema:
        schema_type = "AVRO"
        schema_str = """
        {
          "type": "record",
          "name": "User",
          "fields": [
            {"name": "name", "type": "string"},
            {"name": "age",  "type": "int"}
          ]
        }
        """

    class MockRegisteredSchema:
        schema = MockSchema()

    mock_schema_registry_client.get_latest_version.return_value = MockRegisteredSchema()

    reader = ConfluentRegistryReader(mock_schema_registry_client)
    result = reader.to_recap("dummy_topic")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert result.fields[0].extra_attrs["name"] == "name"
    assert isinstance(result.fields[0], StringType)
    assert result.fields[1].extra_attrs["name"] == "age"
    assert isinstance(result.fields[1], IntType)

    # Check that the get_latest_version method was called with the correct subject.
    mock_schema_registry_client.get_latest_version.assert_called_with(
        "dummy_topic-value"
    )


def test_struct_protobuf():
    mock_schema_registry_client = MagicMock(spec=schema_registry.SchemaRegistryClient)

    class MockSchema:
        schema_type = "PROTOBUF"
        schema_str = """
        syntax = "proto3";

        message Person {
          string name = 1;
          int32 age = 2;
        }
        """

    class MockRegisteredSchema:
        schema = MockSchema()

    mock_schema_registry_client.get_latest_version.return_value = MockRegisteredSchema()

    reader = ConfluentRegistryReader(mock_schema_registry_client)
    result = reader.to_recap("dummy_topic")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert isinstance(result.fields[0], UnionType)
    assert isinstance(result.fields[0].types[1], StringType)
    assert isinstance(result.fields[1], UnionType)
    assert isinstance(result.fields[1].types[1], IntType)

    # Check that the get_latest_version method was called with the correct subject.
    mock_schema_registry_client.get_latest_version.assert_called_with(
        "dummy_topic-value"
    )


def test_struct_json_schema():
    mock_schema_registry_client = MagicMock(spec=schema_registry.SchemaRegistryClient)

    class MockSchema:
        schema_type = "JSON"
        schema_str = """
        {
          "$schema": "http://json-schema.org/draft-07/schema#",
          "title": "User",
          "type": "object",
          "properties": {
            "name": {
              "type": "string"
            },
            "age": {
              "type": "integer"
            }
          },
          "required": ["name", "age"]
        }
        """

    class MockRegisteredSchema:
        schema = MockSchema()

    mock_schema_registry_client.get_latest_version.return_value = MockRegisteredSchema()

    reader = ConfluentRegistryReader(mock_schema_registry_client)
    result = reader.to_recap("dummy_topic")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert result.fields[0].extra_attrs["name"] == "name"
    assert isinstance(result.fields[0], StringType)
    assert result.fields[1].extra_attrs["name"] == "age"
    assert isinstance(result.fields[1], IntType)

    # Check that the get_latest_version method was called with the correct subject.
    mock_schema_registry_client.get_latest_version.assert_called_with(
        "dummy_topic-value"
    )
